#include <asm/io.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include "pemu.h"

#define SHOW_STATUS 0

#define ARGBUFSIZE 0x100000
#define COLUMNS 0x20
#define BUSY_TIMEOUT 10
#define SLEEP_TIMEOUT 1

const char id_string[] =
"MANUFACTURER:Hewlett-Packard;"
"COMMAND SET:HP GDI;"
"MODEL:HP LaserJet 3100;"
"CLASS:PRINTER;"
"DESCRIPTION:Hewlett-Packard LaserJet 3100 MFP;";

static struct pemu *pe;
static struct pemu_log *printer;
static struct proc_dir_entry *proc_status;
static int  inpos = 0;
static char channel_string[2] = " ";

static unsigned char argbuf[ARGBUFSIZE];
static unsigned arglen;
static unsigned char *replay = "\r\nOK\r\n";
static unsigned replaylen = 0;
static void (*next_send)(struct pemu*) = NULL;
static int echo = 1;

static rcount = 0;

static printing=0;

static void
reset(struct pemu *pe)
{
	inpos = 0;
	pe->send = NULL;
	pe->busy_timeout = 0;
	pe->sleep_timeout = 0;
}

static void
dump_buffer(struct pemu *pe, unsigned char *buf, int len)
{
	int column = 0;

	for (; len--; buf++, column++) {

		if (column >= COLUMNS) {
			pemu_log_string(pe->log_high, "\n");
			column = 0;
		}

		if (column)
			pemu_log_string(pe->log_high, " ");

		pemu_log_int(pe->log_high, "%2.2x", *buf);
	}

	if (column)
		pemu_log_string(pe->log_high, "\n");
}

static void
send_echo(struct pemu *pe)
{
	unsigned char echo[] = {0x5a, 0x00, 0x00, 0x00,
				0x04, 0x00, 0x99, 0x99,
				'E', 'c', 'h', 'o'};

	pemu_log_string(pe->log_high, channel_string);
	pemu_log_string(pe->log_high, " <- Echo\n");

	pemu_send(pe, echo, sizeof(echo), 1);
	pe->send = NULL;
}

static unsigned char status[] = {
	0x5a, 0x50, 0x00, 0x00, 0xc4, 0x00, 0x99, 0x99,
	0x34, 0x12, 0x78, 0x56, 0x34, 0x12, 0x78, 0x56,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x29, 0x00, 0x00, 0x00, 0x22, 0x00,
	0x00, 0x00, 0xfb, 0x01, 0x00, 0x00, 0x03, 0x00,
	0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x05, 0x00,
	0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x66, 0x02, 0x6e, 0x7b,
	0x34, 0x12, 0x78, 0x56, 0x34, 0x12, 0x78, 0x56,
	0xfc, 0x0d, 0x01, 0x00, 0xab, 0x00, 0xcd, 0x00,
	0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x42, 0x00,
	0x00, 0x00, 0x00, 0x05, 0x00, 0x00, 0x55, 0x06,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x65, 0x42, 0x65, 0x72, 0x74, 0x69,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x75, 0x41, 0x6f, 0x74, 0x00, 0x00,
	0x80, 0x20, 0x00, 0x00, 0x01, 0x00, 0x00, 0x78,
	0x01, 0x00, 0x00, 0x44, 0x01, 0x00, 0x00, 0x4a,
	0x01, 0x00, 0x00, 0x50};

int
ppstatus_read_proc(char *page, char **start, off_t off,
		   int count, int *eof, void *handle);

static int
pemu_status_write_proc(struct file *file, const char *buffer,
		  unsigned long count, void *data)
{
	struct pemu *pe = (struct pemu*) data;

	if (count != sizeof(status) - 6) {
		pemu_log_int(pe->log_high, "emu3100: count = %4.4x\n", count);
		return -EIO;
	}
	memcpy(status + 6, buffer, count);
	pemu_log_int(pe->log_high, "emu3100: set state %2.2x", (unsigned char) buffer[0]);
	pemu_log_int(pe->log_high, "%2.2x\n", (unsigned char) buffer[1]);
	return count;
}

static void
send_status(struct pemu *pe)
{
	int time = status[0x4d] << 24 | status[0x4c] << 16 | status[0x4f] << 8 | status[0x4e];

	time += 201;

	switch (printing++) {
	case 1:
		break;
	case 2:
		status[0x5d + 8] = 1;
#if 1
		status[0x1a + 8]++;
#endif
		break;
	case 3:
		status[0x5d + 8] = 0;
	default:
		printing = 0;
	}

#if SHOW_STATUS
	pemu_log_string(pe->log_high, channel_string);
	pemu_log_int(pe->log_high, " <- Status (%8.8x,", time);
	pemu_log_int(pe->log_high, "%d)\n", status[0x22]);
#endif

	status[0x4d] = (time >> 24) & 0xff;
	status[0x4c] = (time >> 16) & 0xff;
	status[0x4f] = (time >>  8) & 0xff;
	status[0x4e] = (time      ) & 0xff;

	pemu_send(pe, status, sizeof(status), 1);
	pe->send = NULL;
}

static void
send_version(struct pemu *pe)
{
	unsigned char version[] = {0x5a, 0x50, 0x0b, 0x00,
				   0x04, 0x00, 0x99, 0x99,
				   '.', '1', '2', '1'};

	pemu_log_string(pe->log_high, channel_string);
	pemu_log_string(pe->log_high, " <- VERSION\n");

	pemu_send(pe, version, sizeof(version), 1);
	pe->send = NULL;
}

static void
send_print_ack(struct pemu *pe)
{
	unsigned char echo[] = {0x5a, 0x20, 0x08, 0x00,
				0x00, 0x00, 0x99, 0x99};

	pemu_log_string(pe->log_high, channel_string);
	pemu_log_string(pe->log_high, " <- ACK\n");

	pemu_send(pe, echo, sizeof(echo), 1);
	pe->send = NULL;
}

static void
send_modem_replay(struct pemu *pe)
{
	unsigned char modem_header[] = {0x5a, 0x40, 0x00, 0x00,
					0x00, 0x00, 0x99, 0x99};

	if (!replaylen)
		replaylen = strlen(replay);

	modem_header[4] = (echo ? arglen : 0) + replaylen;

	pemu_log_string(pe->log_high, channel_string);
	pemu_log_int(pe->log_high, " <- MODEM (%x,", replaylen);
	pemu_log_int(pe->log_high, "%x) ", echo);
	pemu_log_string(pe->log_high, replay);
	pemu_log_string(pe->log_high, "\n");

	pemu_send(pe, modem_header, sizeof(modem_header), 1);
	if (echo)
		pemu_send(pe, argbuf, arglen, 1);
	pemu_send(pe, replay, replaylen, 1);
	pe->send = next_send = NULL;
	replaylen = 0;
}

static void
send_modem_firmware(struct pemu *pe)
{
	unsigned char modem_firmwareware[] = {
		0x5a, 0x40, 0x00, 0x00, 0x36, 0x00, 0x99, 0x99,
		'A', 'T', 'I', '3', '\r', '\r', '\n', 'H',
		'e', 'w', 'l', 'e', 't', 't', '-', 'P',
		'a', 'c', 'k', 'a', 'r', 'd', '\n', '\r',
		'H', 'P', ' ', 'L', 'a', 's', 'e', 'r',
		'J', 'e', 't', ' ', '3', '1', '0', '0',
		'\n', '\r', '1', '.', '1', '2', '\n', '\r',
		'\r', '\n', 'O', 'K', '\r', '\n'};

	pemu_log_string(pe->log_high, channel_string);
	pemu_log_string(pe->log_high, " <- MODEM FIRMWARE\n");

	pemu_send(pe, modem_firmwareware, sizeof(modem_firmwareware), 1);
	pe->send = NULL;
}

static void
receive(struct pemu *pe, int data, int status0)
{
	unsigned char modem_conf[] = {0x41,
				      0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b,
				      0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b,
				      0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b, 0x1b,
				      0x0d, 0x0a};

	static unsigned channel, code, unknown;
	static unsigned char cmdbuf[8];

	SET( nFault(H) );

	rcount++;

	if (inpos < 8)
		cmdbuf[inpos] = data;
	else if (inpos - 8 < ARGBUFSIZE)
		argbuf[inpos - 8] = data;
	inpos++;

	switch (inpos) {
	case 1:
		if (cmdbuf[0] == 'Z')
			pe->busy_timeout = jiffies + BUSY_TIMEOUT;
		else {
			pemu_log_int(pe->log_high, "IGNORING %2.2x\n", cmdbuf[0]);
			reset(pe);
		}
		return;

	case 2:
	case 3:
	case 4:
	case 5:
	case 6:
	case 7:
		return;
	case 8:
		channel = cmdbuf[1];
		code    = cmdbuf[2] | (cmdbuf[3] << 8);
		arglen  = cmdbuf[4] | (cmdbuf[5] << 8);
		unknown = cmdbuf[6] | (cmdbuf[7] << 8);

		channel_string[0] = channel ?: ' ';

		if (arglen > ARGBUFSIZE) {
			pemu_log_string(pe->log_high, "COMMAND NAME TOO LONG\n");
			reset(pe);
			return;
		}
	}

	if (inpos - 8 < arglen)
		return;

	if (channel != 0x51 || code != 2 || SHOW_STATUS == 1) {
		pemu_log_int(pe->log_high, "CMD %2.2x ", channel);
		pemu_log_int(pe->log_high, "%4.4x ", code);
		pemu_log_int(pe->log_high, "%4.4x ", arglen);
		pemu_log_int(pe->log_high, "%4.4x    ", unknown);
		pemu_log_int(pe->log_high, "(#%8.8x)\n", rcount);
		dump_buffer(pe, argbuf, arglen);
	}

	reset(pe);

	switch (code) {

	case 0:
		switch (channel) {
		case 0x00:
			if (!arglen) {
				pemu_log_string(pe->log_high, "MISSING ARGUMENT\n");
				return;
			}
			pemu_log_string(pe->log_high, channel_string);
			pemu_log_string(pe->log_high, " -> ");
			pemu_log(pe->log_high, argbuf, arglen);
			pemu_log_string(pe->log_high, "\n");

			if (arglen == 4 && !strncmp(argbuf, "Echo", 4))
				pe->send = send_echo;
			else
				pemu_log_string(pe->log_high, "IGNORED\n");
			break;

		case 0x21:
			printing=1;

			pemu_log_string(pe->log_high, channel_string);
			pemu_log_string(pe->log_high, " -> ");
			pemu_log_string(pe->log_high, "PRINTER DATA\n");
			pemu_log(printer, argbuf, arglen);

			pe->sleep_timeout = jiffies + SLEEP_TIMEOUT;
			break;

		case 0x41:
			if (!arglen) {
				pemu_log_string(pe->log_high, "MISSING ARGUMENT\n");
				return;
			}
			pemu_log_string(pe->log_high, channel_string);
			pemu_log_string(pe->log_high, " -> ");
			pemu_log(pe->log_high, argbuf, arglen);
			pemu_log_string(pe->log_high, "\n");

			if (arglen == 2 && !strncmp(argbuf, "\r\n", 2)) {
			pemu_log_string(pe->log_high, replay);
				if (!next_send) {
					replay = "";
					pe->send = send_modem_replay;
					break;
				}
				pe->send = next_send;
				break;
			}

			replay = "\r\nOK\r\n";
			pe->send = send_modem_replay;

			if (arglen == 5 && !strncmp(argbuf, "\rATZ", 4))
				echo = 1;


			if (arglen == 5 && !strncmp(argbuf, "ATI3", 4))
				pe->send = send_modem_firmware;


			if (arglen == 0x1b && !strncmp(argbuf, modem_conf, 0x1b)) {
				replay = "";
			}

			if (arglen == 2 && !strncmp(argbuf, "UU", 2)) {
				pe->send = send_modem_replay;
				replay = "HP LaserJet 3100\0"
					"1.12\0"
					"26 Mar'98 17:45\0";
				replaylen = 17 + 5 + 16;
				echo = 0;
				break;
			}
			break;

		default:
			pemu_log_string(pe->log_high, "UNEXPECTED CHANNEL\n");
		}
		break;

	case 1:
		pemu_log_string(pe->log_high, channel_string);
		pemu_log_string(pe->log_high, " -> INIT\n");
		if (channel || unknown || arglen)
			pemu_log_string(pe->log_high, "UNEXPECTED PARAMETER\n");
		break;


	case 2:
		if (channel != 0x51 || code != 2 || SHOW_STATUS == 1) {
			pemu_log_string(pe->log_high, channel_string);
			pemu_log_string(pe->log_high, " -> REQUEST STATUS\n");
		}
		if (channel != 'Q' || unknown || arglen)
			pemu_log_string(pe->log_high, "UNEXPECTED PARAMETER\n");
		pe->send = send_status;
		break;

	case 9:
		pemu_log_string(pe->log_high, channel_string);
		pemu_log_string(pe->log_high, " -> REQUEST PRINT\n");
		if (channel != 0x21 || unknown || arglen)
			pemu_log_string(pe->log_high, "UNEXPECTED PARAMETER\n");
		pe->send = send_print_ack;
		break;

	case 0xb:
		pemu_log_string(pe->log_high, channel_string);
		pemu_log_string(pe->log_high, " -> REQUEST VERSION\n");
		if (channel != 'Q' || unknown || arglen)
			pemu_log_string(pe->log_high, "UNEXPECTED PARAMETER\n");
		pe->send = send_version;
		break;

	default:
		pemu_log_int(pe->log_high, "UNKNOWN CMD %2.2x ", channel);
		pemu_log_int(pe->log_high, "%4.4x ", code);
		pemu_log_int(pe->log_high, "%4.4x ", arglen);
		pemu_log_int(pe->log_high, "%4.4x\n", unknown);
		dump_buffer(pe, argbuf, arglen);
	}

	if (pe->send) {
		if (channel != 0x51 || code != 2 || SHOW_STATUS == 1)
			pemu_log_string(pe->log_high, "nFault(L)\n");
		SET( nFault(L) );
		pe->busy_timeout = jiffies + BUSY_TIMEOUT;
	}
}

#ifdef MODULE
int
init_module(void)
{
	pe = new_pemu("lj3100");
	if (!pe)
		return -ENOMEM;

	printer = new_pemu_log(pe, "printer");
	if (!printer) {
		del_pemu(pe);
		return -ENOMEM;
	}

	pe->negotiate = NULL;
	pe->receive   = receive;
	pe->send      = NULL;
	pe->id_string = id_string;


	proc_status = new_proc_entry("status", S_IFREG | S_IRUGO | S_IWUGO, pe->proc_dir, 0, NULL);
	proc_status->write_proc = pemu_status_write_proc;
	proc_status->read_proc = ppstatus_read_proc;
	proc_status->data = pe;

	reset(pe);

	return 0;
}

void
cleanup_module(void)
{
	if (proc_status)
		destroy_proc_entry(pe->proc_dir, &proc_status);
	del_pemu_log(printer);
	del_pemu(pe);
}
#endif
