#include <stdio.h>

static unsigned column = 0;
static int olddir = 0;

void dump_byte(int dir, unsigned data)
{
	if (column && dir != olddir) {
		printf("\n");
		column = 0;
	}

	olddir = dir;

	printf("%s%2.2x%s",
	       column     == 0 ? (dir ? "\t" : "#\t") :
	       column % 8 == 0 ? "   " :
	       column % 2 == 0 ? " " : "",

	       data,

	       column == 15 ? "\n" : "");

	column++;
	column %= 16;
}

void dump_header(int dir, unsigned channel, unsigned code, unsigned length, unsigned flags)
{
	printf(column ? "\n\n" : "\n");
	printf("%s CMD %4x %4x %4.4x %4.4x\n",
	       dir > 0 ? "-> " : "<- ",
	       channel, code, length, flags);
	column = 0;
}

int main()
{
	char ch;
	int pos = 0;

	for (;;) {
		unsigned count, status, data, channel, code, length, flags, dir;

		if (!pos)
			channel = code = length = flags = dir = 0;

		if (scanf(" BREAK%c ", &ch) == 1) {
			if (feof(stdin)) {
				printf(column ? "\n\n" : "\n");
				return 0;
			}
			printf("\nBREAK\n");
			pos = 0;
			column = 0;
		}
		else if (scanf("%4x %2x%2x\n", &count, &status, &data) == 3) {
			if (status == 0x30)
				dir = 1;
			else if (status == 0x48)
				dir = -1;
			else
				continue;

			switch (pos++) {
			case 0:
				if (data == 0x5a)
					break;
				pos = 0;
				dump_byte(0, data);
				break;

			case 1:
				channel = data;
				break;

			case 2:
				code = data;
				break;

			case 3:
				code |= data << 8;
				break;

			case 4:
				length = data;
				break;

			case 5:
				length |= data << 8;
				break;

			case 6:
				flags = data;
				break;

			case 7:
				flags |= data << 8;
				dump_header(dir, channel, code, length, flags);
				if (length > 0x400) {
					printf("oversized packet\n");
					length = 0;
				}
				if (!length)
					pos = 0;
				break;

			default:
				dump_byte(dir, data);
				if (--length == 0)
					pos = 0;
			}
		}
		else {
			if (feof(stdin)) {
				printf(column ? "\n\n" : "\n");
				return 0;
			}
			else {
				fprintf(stderr, "format error count=%x status=%x data=%x\n", count, status, data);
				return -1;
			}
		}
	}
}
