#ifndef PEMU_H
#define PEMU_H

#define PORT0           0x278
#define PORT1           0x378

#define IRQ0            7

/* Cable definition (port, xname)              remote    local */
#define X_Busy          0, nAutoFd          /*   11   <-   14  */
#define X_nAck          0, nStrobe          /*   10   <-    1  */
#define X_PError        0, nInit            /*   12   <-   16  */
#define X_Select        1, nSelectIn        /*   13   <-   17 (2nd port) */
#define X_nFault        0, nSelectIn        /*   15   <-   17  */
#define X_nSelectIn     0, nFault           /*   17   ->   15  */
#define X_nInit         0, PError           /*   16   ->   12  */
#define X_nAutoFd       0, Busy             /*   14   ->   11  */
#define X_nStrobe       0, nAck             /*    1   ->   10  */ /* hardcoded, do not change */
/* Data */           /* 0 */                /*  2,9   <->  2,9 */
/* Ground */         /* 0,1 */              /*   25   <->  25  */

#define REJECT_1284    -1
#define ACCEPT_1284     0
#define NO_DATA_1284    1

#define NIBBLE_MODE     0
#define ECP_MODE     0x10
#define ID_MODE_FLAG    4

/* ioports */
#define	DATA(portnum)		(PORT##portnum + 0)
#define	WSTATUS(portnum)	(PORT##portnum + 0)
#define	STATUS(portnum)		(PORT##portnum + 1)
#define	CONTROL(portnum)	(PORT##portnum + 2)
#define EPPADDR(portnum)	(PORT##portnum + 3)
#define EPPDATA(portnum)	(PORT##portnum + 4)
#define ECPA(portnum)		(PORT##portnum + 0x400)
#define ECPB(portnum)		(PORT##portnum + 0x401)
#define ECONTROL(portnum)	(PORT##portnum + 0x402)

/* EControl register */
#define MODE_PS2		(1 << 5)
#define ERRINT_DISABLE		(1 << 4)
#define DMA_DISABLE		(0 << 3)
#define SERVINT_DISABLE		(1 << 2)
/* Control register */
#define DIRECTION_IN		(1 << 5)
#define DIRECTION_OUT		(0 << 5)
#define ACKINT			(1 << 4)
#define ACKINT_ENABLE		(1 << 4)
#define ACKINT_DISABLE		(0 << 4)
#define HW_OUT_MASK		(0xf << 0)

/* I/O bits             (bit, invert) */
#define HW_Busy			7, 1  /* Status register */
#define HW_nAck			6, 0
#define HW_PError		5, 0
#define HW_Select		4, 0
#define HW_nFault		3, 0
#define HW_nSelectIn		3, 1  /* Control register*/
#define HW_nInit		2, 0
#define HW_nAutoFd		1, 1
#define HW_nStrobe		0, 1

/* Synonymes for nibble and byte mode */
#define PtrBusy			Busy
#define PtrClk			nAck
#define AckDataReq		PError
#define Xflag			Select
#define nDataAvail		nFault
#define Active1284		nSelectIn
#define HostBusy		nAutoFd
#define HostClk			nStrobe

/* Synonymes for ECP mode */
#define PeriphAck		Busy
#define PeriphClk		nAck
#define nAckReverse		PError
#define nPeriphRequest		nFault
#define nReverseRequest		nInit
#define HostAck			nAutoFd

#define Busy(HLX)		HLX(Busy)
#define nAck(HLX)		HLX(nAck)
#define PError(HLX)		HLX(PError)
#define Select(HLX)		HLX(Select)
#define nFault(HLX)		HLX(nFault)
#define nSelectIn(HLX)		HLX(nSelectIn)
#define nInit(HLX)		HLX(nInit)
#define nAutoFd(HLX)		HLX(nAutoFd)
#define nStrobe(HLX)		HLX(nStrobe)

#define H(name)                       HLX1(X_##name, 1, 1)
#define L(name)                       HLX1(X_##name, 1, 0)
#define X(name)                       HLX1(X_##name, 0, 1)
#define HLX1(X_name, m, v)            HLX2(X_name, m, v)
#define HLX2(portnum, xname, m, v)    HLX3(portnum, HW_##xname, m, v)
#define HLX3(portnum, HW_xname, m, v) HLX4(portnum, HW_xname, m, v)
#define HLX4(portnum, bit, inv, m, v) ((m << (16 + bit) | (v ^ (inv & m)) << bit) << (portnum * 8))

#define MASK(portnum, flags)          (((flags) >> (16 + portnum * 8)) & 0xff)
#define VAL(portnum, flags)           (((flags) >> (portnum * 8)) & 0xff)

#define CHECK(flags)                  CHECKV(inb(STATUS(0)), inb(STATUS(1)), flags)
#define CHECKV(in0, in1, flags)       (CHECKV1(in0, MASK(0, (flags)), VAL(0, (flags))) &&\
                                       CHECKV1(in1, MASK(1, (flags)), VAL(1, (flags))))
#define CHECKV1(in, mask, val)        (!mask || ((in & mask) == val))

#define SET(flags)                    do {\
                                              SET1(CONTROL(1), MASK(1, (flags)), VAL(1, (flags)));\
                                              SET1(CONTROL(0), MASK(0, (flags)), VAL(0, (flags)));\
                                      } while(0)
#define SET1(port, mask, val)         if (mask || val)\
                                              outb(((inb(port) & ~(mask & HW_OUT_MASK)) ^ (val & HW_OUT_MASK)), port)

#define SETC(portnum, val0, flags)    SETC1(CONTROL(portnum), (val0), MASK(portnum, (flags)), VAL(portnum, (flags)))
#define SETC1(port, val0, mask, val)  outb(((val0 & ~(mask & HW_OUT_MASK)) ^ (val & HW_OUT_MASK)), port)

#define GETV(portnum, in, flags)      GETV1(in, MASK((portnum), (flags)), VAL((portnum), (flags)))
#define GETV1(in, mask, val)          (((in) & ~(mask & HW_OUT_MASK)) ^ (val & HW_OUT_MASK))

#define REPEAT(num, x)		REPEAT1(num, x)
#define REPEAT1(num, x)		REPEAT_##num(x)
#define REPEAT_2(x)		x x
#define REPEAT_4(x)		REPEAT_2(x)   REPEAT_2(x)
#define REPEAT_8(x)		REPEAT_4(x)   REPEAT_4(x)
#define REPEAT_16(x)		REPEAT_8(x)   REPEAT_8(x)
#define REPEAT_32(x)		REPEAT_16(x)  REPEAT_16(x)
#define REPEAT_64(x)		REPEAT_32(x)  REPEAT_32(x)
#define REPEAT_128(x)		REPEAT_64(x)  REPEAT_64(x)

struct pemu;
struct pemu_buffer;
struct proc_dir_entry;

struct pemu {
	int mode;
	unsigned long busy_timeout;
	unsigned long sleep_timeout;
	struct proc_dir_entry *proc_dir;
	struct proc_dir_entry *proc_run;
	struct proc_dir_entry *proc_status;
	struct pemu_log       *log_low;
	struct pemu_log       *log_high;

	const char *id_string;
	int  (*negotiate)(struct pemu*, int mode);
	void (*receive)(struct pemu*, int data, int status);
	void (*send)(struct pemu*);

	void *data;
};

#define BUFFERSIZE (0x8000 - 2 * sizeof(long) - 2 * sizeof(struct emu_buffer*))
struct pemu_buffer {
	struct pemu_buffer *prev;
	struct pemu_buffer *next;
	long pos;
	int buffer_full;
	char buf[BUFFERSIZE];
};

struct pemu_log {
	struct pemu *pe;
	struct pemu_buffer *buffer_list;
	spinlock_t buffer_lock;
	struct proc_dir_entry *proc_log;
};

struct pemu *new_pemu(const char *dirname);
void del_pemu(struct pemu *pe);

void pemu_setup(struct pemu*);
void pemu_cleanup(struct pemu*);

int pemu_send(struct pemu*, const unsigned char *buf, size_t nbytes, int eof);
int pemu_send_id(struct pemu*, const char *id);

struct pemu_log *new_pemu_log(struct pemu*, const char *name);
void del_pemu_log(struct pemu_log*);
int pemu_log(struct pemu_log*, const char * buf, size_t nbytes);
int pemu_log_string(struct pemu_log*, const char * string);
int pemu_log_int(struct pemu_log*, const char * format, int val);

struct proc_dir_entry *
new_proc_entry(const char *name, mode_t mode,
	       struct proc_dir_entry *parent,
	       unsigned short ino,
	       struct inode_operations *ops);
void
destroy_proc_entry(struct proc_dir_entry *root, 
		   struct proc_dir_entry **d);
#endif PEMU_H
